from Models import Project, Architecture, Executable, Project
from Models import Workflow, WorkflowStatus
from Models import WorkflowEntry, EntryStatus, EntryPriorityHint, PhysicsGroup, IdToPhysicsGroup, ConsumerName
from Models import Payload
from Models import User
from Models import ApprovedRequest
from Models import Location
from Models import Directory, DirectoryEntry
from Models import GeneratorInfo
from Models import SpecInfo
from Models.RequestProxy import RequestProxy
from ProdRequestCommon.PersistencyTools import Serializer, Serializable
from ProdRequestCommon.JSON import PythonDictSerializer as PythonDictSerializerJSON

from Tools.Functors import AlwaysFalse

from Framework import rpccall, xmlBooleanResponse, templatepage, templateXML, exposeSerialized
from Framework import Controller
from Framework.Controller import require_args
from Framework.Utilities import selectFromDict
from Framework.Application import Application as FrameworkApplication
from cherrypy import expose
import cherrypy
from sqlobject import *
from Tools.SecurityModuleCore import is_authorized, is_authenticated, has_role
from Tools.SecurityModuleCore import Role, Group, SecurityToken
from Tools.SecurityModuleCore import check_ownership

from Tools.Functors import FetchFromArgs

from Modules.Utilities.Serializers.XML import PythonListSerializer, SQLObjectSerializer
from ProdRequestCommon.DataBase import clear_cache

from Framework.PluginManager import DeclarePlugin
from ProdRequestCore import ProdRequestCore
from ProdRequestCfg import ProdRequestCfg
from RequestPolicies import OperatorsRequest, ModifiableRequest
from Tools import catch_exception

from ProdRequestCommon.MonitoringCore import MonitoringCore
from ProdRequestCommon import declare_arg_converter

def defaultProdCommon (cmsswVersion):
    try:
        name, major, minor, revision = cmsswVersion.split ("_", 3)
        if int (major) == 1 and int (minor) < 4:
            return "PRODCOMMON_0_3_X"
        elif int (major) == 1 and int (minor) == 5:
            return "PRODCOMMON_0_3_X_SLC4"
        elif int (major) == 1 and int (minor) == 6:
	        return "PRODCOMMON_0_3_X_SLC4" 
        elif int (major) == 1 and int (minor) == 7:
            return "PRODCOMMON_0_3_X_SLC4"
	else:
            return "PRODCOMMON_0_3_X_SLC4"
    except:
        return "PRODCOMMON_0_3_X_SLC4"

class RedirectorToLogin (object):
    def __init__ (self, page):
        self.page = page
    def __call__ (self, controller):
        return controller.templatePage ("gotoLogin", {'requestedPage': self.page})

class CfgTreeBrowser (Controller):
    def __init__ (self, context):
        Controller.__init__ (self, context, __file__)
    
    @exposeSerialized (serializer=SQLObjectSerializer ("cfgs", "cfg"))
    def browse (self, branchName):
        twigs = Directory.select (Directory.q.name == branchName)
        for twig in twigs:
            return twig.nodes
        return {'cfgs': []}
        
DeclarePlugin ("/Controllers/ProdRequest/CfgTreeBrowser", CfgTreeBrowser, options={"baseUrl": "/ProdRequest/CfgTreeBrowser"})



    
def ref_workflow_entry_by_id (func):
    def getWorkflowEntryById (entry_id):
        return WorkflowEntry.get (int (entry_id))
    return declare_arg_converter ('entry', '_id', getWorkflowEntryById)(func)

def ref_project_by_id (func):
    def getProjectById (project_id):
        return Project.get (int (project_id))
    return declare_arg_converter ('project', '_id', getProjectById)(func)

def ref_project_by_name (func):
    def getProjectByName (project_name):
        projects = Project.select (Project.q.name == project_name)
        if not projects.count ():
            return None
        return projects[0]
    return declare_arg_converter ('project', '_name', getProjectByName)(func) 

def ref_location_by_id (func):
    def getLocationById (location_id):
        return Location.get (int (location_id))
    return declare_arg_converter ('location', '_id', getLocationById)(func)

def ref_workflow_by_id (func):
    return declare_arg_converter ('workflow', '_id', Workflow.get)(func)

def ref_payload_by_id (argName):
    def getPayloadById (payload_id):
        payload_id = int (payload_id)
        return (payload_id and Payload.get (payload_id)) or None
    def decorator (func):
        wrapper = declare_arg_converter (argName, '_id', getPayloadById)(func)
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper
    return decorator
    
#TODO: separate the DB maintainance API from the webpage generation
class ProdRequest(Controller):
    def __init__ (self, context):
        Controller.__init__ (self, context, __file__)
        self.core = ProdRequestCore ()
        self.monitoring = MonitoringCore (context)
        self.cfg = ProdRequestCfg (context)
        self.context.addService (self.cfg)
        
    def readyToRun (self):
        self.context.Logger ().trace ("Creating ProdRequest roles and groups")
        securityApi = self.context.SecurityDBApi ()
        securityApi.api.addRole ("Production Manager")
        securityApi.api.addRole ("Production Operator")
        securityApi.api.addRole ("Admin")
        securityApi.api.addGroup ("ProdRequest")
        
        self.core.createDBSchema ()
        self.cfg.initialise ()
        opts, args = self.context.OptionParser ().parse_args ()
        self.baseUrl = opts.baseUrl
        
        # Get the plugin manager and instanciate as many ProdCommonPlugins as required
        plugins = self.context.PluginManager ().plugins ("/ProdRequest/ProdCommonPlugins")
        self.prodCommonPlugins = {}
        for plugin in plugins:
            self.context.Logger ().trace ("Loading ProdCommonPlugin: %s" % plugin.implementation.__name__)
            prodCommonPlugin = plugin.createInstance (self.context, plugin.options)
            self.prodCommonPlugins[plugin.name.rsplit ("/",1)[1]] = prodCommonPlugin
        
    @expose
    def index (self):
        """Returns the version of the module."""
        return "ProdRequest v0.1"
        
    @templatepage
    def initSession (self):
        return {'redirect': "../ProdRequest/getHome"}

    @templatepage
    def createNewWorkflow (self, **args):
        w = Workflow (name="Dummy name")
        return {'workflow': w.id}

    #TODO: fix urls
    @expose
    @is_authorized (Role('Admin'), Group('ProdRequest'), onFail=RedirectorToLogin ("../ProdRequest/getAdministrationPage"))
    def getAdministrationPage (self, **args):
        return self.templatePage ("getAdministrationPage", {'dbsUrl': "",
                                                            'dbsInstance': ""})

    @expose
    @is_authenticated (onFail=RedirectorToLogin ("../ProdRequest/createNewWorkflow"))
    def getPage (self, **args):
        currentWorkflow=1
        if args.has_key ("selected_workflow"):
            currentWorkflow = int (args["selected_workflow"])

        editedWorkflow=currentWorkflow
        if args.has_key ("editedWorkflow"):
            editedWorkflow=args["editedWorkflow"]
        
        currentPayload=1
        if args.has_key ("selected_payload"):
            currentPayload = int (args["selected_payload"])
        
        projectSelect = Project.select ()
        architectureSelect = Architecture.select ()
        locationSelect = Location.select ()

        #FIXME: rename payloads to selectedPayloads
        #FIXME: rename workflowsEntries to editedPayloads
        return self.templatePage ("getPage", {'editedWorkflow': editedWorkflow,
                'currentWorkflow': currentWorkflow,
                'currentPayload': currentPayload,
                'architectures': architectureSelect,
                'projects': projectSelect,
                'locations': locationSelect})

    @xmlBooleanResponse
    def createPayloadWithEntry (self, 
                                name=None,
                                submit=None, 
                                requested_events=1000, 
                                physics_group="Individual", 
                                selection_efficiency=1.0,
                                cross_section=0,
                                output_dataset=None,
                                primary_dataset=None,
                                input_dataset=None,
                                pileup_dataset=None,
                                parameterset=None,
                                prodcommon_version=None,
                                version=None,
                                request_type="cmsRun",
                                generator_name=None,
                                generator_executable=None,
                                generator_cfg=None,
                                use_prodmanager="0",
                                **args):
        if version:
            projects = Project.select (Project.q.name == version)
            if not projects.count ():
                self.context.Logger ().message ("CMSSW version not available %s" % version)
                return False
            project = projects[0] 
            version = project.name
        else:
            project = Project.get (1)            
            version = project.name
                                         
        workflowEntryArgs={}
        
        consumer = (use_prodmanager == "1") and ConsumerName.PRODMANAGER or ConsumerName.OPERATOR  
        primary_dataset = primary_dataset or name.rsplit (".", 1)[0]
        prodcommon_version = prodcommon_version or defaultProdCommon (version)
                
        currentWorkflowId=0
        if args.has_key ("editedWorkflow"):
            currentWorkflowId = args["editedWorkflow"]
        elif args.has_key ("workflowName"):
            workflowName = args["workflowName"]
            workflowSelect = Workflow.select (Workflow.q.name == workflowName)
            if workflowSelect.count ():
                for workflow in workflowSelect:
                    currentWorkflowId = workflow.id
            else:
                dn = SecurityToken ().dn
                workflow = dn and Workflow (name=workflowName) or Workflow (name=workflowName, 
                                                                            owner_id=0)
                currentWorkflowId = workflow.id
        else:
            raise "Missing parameter"
        payload = Payload (name=name, parameterset=parameterset, version=version)
        status = submit and EntryStatus.PENDING or EntryStatus.INJECTED
        try:
            physics_group=PhysicsGroup.__getattr__ (physics_group)
        except KeyError:
            physics_group=0

        # If the request if of type "cmsGen", create an entry in the 
        # GeneratorInfo table and associate it to the current request.
        generatorInfo = None
        if request_type == "cmsGen":
            if "" in (generator_name, generator_cfg):
                return False
            generatorInfo = GeneratorInfo (generator=generator_name,
                                           executable="",
                                           cfg=generator_cfg)
        
        WorkflowEntry (name=payload.name,
                       payload=payload,
                       workflowID=currentWorkflowId,
                       status=status,
                       architecture=None,
                       project=project,
                       requested_events=int (requested_events),
                       physics_group=physics_group,
                       selection_efficiency=float (selection_efficiency),
                       cross_section=float (cross_section),
                       output_dataset=output_dataset,
                       input_dataset=input_dataset,
                       primary_dataset=primary_dataset,
                       prodcommon_version=prodcommon_version,
                       generator_info=generatorInfo,
                       preferred_consumer=consumer,
                       **workflowEntryArgs)
        return True
    
    #TODO: check if user can create a payload
    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    def createPayloadNode (self, **args):
        creationArgs = selectFromDict (args, "name",
                                             "parameterset",
                                             "version")
        Payload (**creationArgs)
        return True
        
    #TODO: rename to just createWorkflow
    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    @require_args ("name", onFail=AlwaysFalse)
    def createWorkflowSpec (self, name, **args):
        Workflow (name=name)
        return True
    
    #TODO: check if user can add to the given Workflow
    #FIXME: name is not a property that should belong to a WorkflowEntry!!!
    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    @ref_workflow_by_id
    @ref_payload_by_id ('payload')
    @ref_payload_by_id ('payloadprev')
    @ref_payload_by_id ('payloadparent')
    @require_args ("workflow", "payload", "payloadprev", "payloadparent", onFail=AlwaysFalse)
    def createWorkflowSpecEntry (self, payload=None, payloadprev=None, payloadparent=None, workflow=None, **args):
        name = payload.name.rsplit (".", 1)[0]
        project = Project.get (1)
        WorkflowEntry (workflow=workflow, 
                       payload=payload,
                       payloadprev=payloadprev,
                       payloadparent=payloadparent,
                       name=name,
                       status=EntryStatus.INJECTED,
                       project=project,
                       prodcommon_version=defaultProdCommon (project.name),
                       primary_dataset=args.get ("primary_dataset", "None"),
                       output_dataset=args.get ("output_dataset", "Click to edit"),
                       pileup_dataset=args.get ("pileup_dataset", "None"))
        return True
    
    #TODO: change application_id to project_id, when Request is moved to have projects, 
    #rather than applications
    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    @ref_workflow_by_id
    @ref_location_by_id
    @require_args ("workflow", "location")
    @catch_exception (SQLObjectNotFound)
    def submitWorkflow (self, workflow=None, location=None, **args):
        creationArgs = selectFromDict (args, "architecture_id",
                                             "comment")
        for entry in workflow.entries:
            if entry.primary_dataset in (None, "None", "undefined"):
                return False
            entry.addLocation (location)
            entry.status = EntryStatus.PENDING
        return True
    
    @exposeSerialized (serializer=SQLObjectSerializer ("architectures", "architecture"))
    @clear_cache (Architecture)
    def getAvailableArchitectures (self, **args):
        """Returns a list of the available architectures"""
        return Architecture.select ()
    
    @exposeSerialized (serializer=SQLObjectSerializer ("executables", "executable"))
    @clear_cache (Executable)
    def getAvailableExecutables (self, **args):
        return Executable.select ()

    @exposeSerialized (serializer=SQLObjectSerializer ("projects", "project"))
    @clear_cache (Project)
    def getAvailableProjects (self, **args):
        return Project.select ()

    @expose
    @has_role (Role ("Production Manager"), onFail=RedirectorToLogin ("../ProdRequest/getApprovalPage"))
    def getApprovalPage (self, **args):
        raise cherrypy.HTTPRedirect (self.baseUrl + "/ProdRequest/getMonitoringPage")
        return "You should not see this message. If you do, please file a bug report."

    def userMonitoringPage (self, **args):
        self.context.Logger ().message ("You are viewing casual user view")
        return self.templatePage ("userMonitoringPage", {})

    #TODO: use the "region" information to provide only the workflow associated to a given team.
    # FIXME: APPROVED_PRODMANAGER requests should be shown but download link should point to prodmanager.
    @has_role (Role ("Production Operator"), onFail=lambda self: self.userMonitoringPage())
    def productionOperatorPage (self, **args):
        self.context.Logger ().message ("You are viewing production operator view")
        return self.templatePage ("productionOperatorPage", {'pendingRequests': [],
                'statuses': [("Approved", EntryStatus.APPROVED),
                             ("Started", EntryStatus.RUNNING),
                             ("Done", EntryStatus.DONE)]})
        
    @expose
    @has_role (Role ("Production Manager"), onFail=lambda self : self.productionOperatorPage())
    def getMonitoringPage (self, **args):
        self.context.Logger ().message ("You are viewing the manager view")
        return self.templatePage ("getMonitoringPage", {'pendingRequests': [],
                'statuses': [("Pending", EntryStatus.PENDING),
                             ("Approved", EntryStatus.APPROVED),
                             ("Approved (ProdManager)", EntryStatus.APPROVED_PRODMANAGER),
                             ("Started", EntryStatus.RUNNING),
                             ("Done", EntryStatus.DONE),
                             ("Deleted", EntryStatus.RETIRED),
                             ("Rejected", EntryStatus.REJECTED)],
                'priorities': [("Low", 20),
                               ("Normal", 50),
                               ("High", 80)],
                'locations': Location.select ()})
        
        
    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    @require_args ("source_workflow", "dest_workflow")
    def addFromWorkflow (self, source_workflow, dest_workflow, **args):
        """ Adds a WorkflowEntry found in source_workflow in dest_workflow.
        """
        entries = WorkflowEntry.select (WorkflowEntry.q.workflow_id == source_workflow)
        for entry in entries:
            WorkflowEntry (name=entry.name,
                           status=0,
                           workflow_id=dest_workflow,
                           payload_id=entry.payload_id,
                           payloadprev_id=0,
                           payloadparent_id=0,
                           primary_dataset=entry.primary_dataset,
                           output_dataset=entry.output_dataset,
                           pileup_dataset=entry.pileup_dataset,
                           prodcommon_version=defaultProdCommon (entry.project.name),
                           project=entry.project)
        return True
    
    @clear_cache (WorkflowEntry)
    def getRequestsByStatus (self, *stati, **kwds):
        conditions = []
            
        for status in stati:
            conditions.append (WorkflowEntry.q.status == status)
        security_api = self.context.SecurityDBApi ()
        token = SecurityToken ()
        if not len (security_api.groupsForRole (token.dn, "Production Manager")):
            requestCondition = AND (OR(*conditions), 
                Workflow.q.id == WorkflowEntry.q.workflowID,
                Workflow.q.owner_id == SecurityToken ().userId)
        else:
            requestCondition = OR (*conditions)
        
        if kwds.has_key ("filter_name"):
            filterCondition = LIKE(WorkflowEntry.q.primary_dataset, "%%%(filter_name)s%%" % kwds)
            requestCondition = AND (filterCondition, requestCondition)
        
        entries = WorkflowEntry.select (requestCondition)
        results = []
        securityApi = self.context.SecurityDBApi ()
        for entry in entries:
            try:
                user=security.getUsernameFromID (securityApi.entry.workflow.owner_id)
            except:
                user="Unknown"
            
            #Should really use joins...
            results.append ({"name": entry.name,
                             "user": user,
                             "id": entry.id,
                             "workflow_id": entry.workflow.id,
                             "events": entry.requested_events,
                             "status_id": entry.status,
                             "comment": entry.workflow.comment,
                             "location_hints": entry.location_hints,
                             "priority": entry.priority_hint,
                             "entry": entry})
        return results
    
    def paginateResults (self, entries, page=0, entriesPerPage=-1):
        page = int (page)
        entriesPerPage = int (entriesPerPage)
        initialRequest = entriesPerPage * page
        finalRequest = entriesPerPage * (page + 1)
        if len (entries) < finalRequest:
            finalRequest = len (entries)
        if entriesPerPage < 0:
            finalRequest = len (entries)
        return {"initialRequest": initialRequest,
                "finalRequest": finalRequest,
                "totalRequests": len (entries),
                "requests": entries[initialRequest:finalRequest]}
    
    @templateXML
    @ref_location_by_id
    @require_args ("location")
    def getRequestsByLocation (self, location=None, page=0, **args):
        entries = location.entries.filter (OR (WorkflowEntry.q.status == EntryStatus.APPROVED, 
                                               WorkflowEntry.q.status == EntryStatus.APPROVED_PRODMANAGER))
        results = []
        for entry in entries:
            try:
                user = self.context.SecurityDBApi.getUsernameFromID (entry.workflow.owner_id)
            except:
                user = "Unknown"
                
            #Should really use joins...
            results.append ({"name": entry.name,
                             "user": user,
                             "id": entry.id,
                             "workflow_id": entry.workflow.id,
                             "events": entry.requested_events,
                             "status_id": entry.status,
                             "comment": entry.workflow.comment,
                             "location_hints": entry.location_hints,
                             "priority": entry.priority_hint})
        return self.paginateResults (results, page, entriesPerPage = self.cfg.entriesPerPage)
    
    @expose
    @require_args ("request_id", onFail=str)
    def getRequestCfgFile (self, request_id, page = 0):
        workflow = WorkflowEntry.get (int (request_id))
        return workflow.payload.parameterset
        
    @templateXML
    @clear_cache (WorkflowEntry)
    def getApprovedRequests (self, **args):
        pendingRequests = self.getRequestsByStatus (EntryStatus.APPROVED, 
                                                    EntryStatus.LOCALSUBMIT)
        page=int (args["page"])
        initialRequest = self.cfg.entriesPerPage * page
        finalRequest = self.cfg.entriesPerPage * (page + 1)
        if len (pendingRequests) < finalRequest:
            finalRequest = len (pendingRequests)
        return {"initialRequest": initialRequest,
                "finalRequest": finalRequest,
                "totalRequests": len(pendingRequests) ,
                "requests": pendingRequests[initialRequest:finalRequest]}
    
    @templateXML
    @clear_cache (WorkflowEntry)
    def getAllRequests (self, page=0, sortBy=None, **args):
        filterRule = None
        if args.has_key ("filter_name"):
            filterRule = LIKE(WorkflowEntry.q.primary_dataset, "%%%(filter_name)s%%" % args)
        
        if filterRule:
            entries = WorkflowEntry.select (AND (WorkflowEntry.q.status != EntryStatus.INJECTED, 
                                            filterRule))
        else:
            entries = WorkflowEntry.select (WorkflowEntry.q.status != EntryStatus.INJECTED)
            
        requests = []
        
        def sortByPriority (x, y):
            return y["entry"].priority_hint - x["entry"].priority_hint

        sortingMap = {"priority": sortByPriority}

        securityApi = self.context.SecurityDBApi ()
        for entry in entries:
            try:
                user=securityApi.getUsernameFromID (entry.workflow.owner_id)
            except:
                user="Unknown"
            
            #Should really use joins...
            requests.append ({"entry": entry,
                             "user": user})
        if sortingMap.has_key (sortBy):
            requests.sort (sortByPriority)
        return self.paginateResults (requests, page=page, entriesPerPage = self.cfg.entriesPerPage)

    @expose
    @clear_cache (WorkflowEntry)
    def getAllRequestsJSON (self, page=0, sortBy=None, **args):
        #try:
        #    WorkflowEntry.get (1).spec_info
        #except:
        #    WorkflowEntry._connection.queryAll ("""alter table workflow_entry add column spec_info_id;""")
        
        #convertibleEntries = WorkflowEntry.select (AND (WorkflowEntry.q.status != EntryStatus.INJECTED,
        #                                                WorkflowEntry.q.request_spec != ""))
        #for entry in convertibleEntries:
        #    entry.spec_info = SpecInfo (text=entry.request_spec)
        #    entry.request_spec = ""

        filterRule = None
        
        if "filter_name" in args:
            filterRule = LIKE(WorkflowEntry.q.primary_dataset, "%%%(filter_name)s%%" % args)
        if filterRule:
            entries = WorkflowEntry.select (AND (WorkflowEntry.q.status != EntryStatus.INJECTED, 
                                            filterRule))
        else:
            entries = WorkflowEntry.select (WorkflowEntry.q.status != EntryStatus.INJECTED)
        
    
        def sortByPriority (x, y):
            return y.priority_hint - x.priority_hint
        
        def getUsername (entry):
            try:
                return securityApi.getUsernameFromID (entry.workflow.owner_id)
            except:
                return "Unknown"
            
        sortingMap = {"priority": sortByPriority}
        
        securityApi = self.context.SecurityDBApi ()
        proxySerializer = Serializer (RequestProxy)

        locationList = [ x for x in Location._connection.queryAll ("""
SELECT workflow_entry_id, location.name 
FROM location_workflow_entry, location 
WHERE location.id = location_workflow_entry.location_id;""")]
        custLocations = {}
        for e, loc in locationList:
            custLocations.setdefault (e, []).append (loc)

        requests = [ RequestProxy (e, getUsername (e), custLocations)
                     for e in entries]
        if sortingMap.has_key (sortBy):
            requests.sort (sortByPriority)
        results = {"Requests": {"totalItems" : len (requests),
                                "itemsFound": len (requests),
                                "items": [ Serializable (r, proxySerializer)
                                           for r in requests ]}}
        return PythonDictSerializerJSON () (results)
    
    @templateXML
    def getUserRequests (self, page=0, filter_name="", **args):
        requests = self.getRequestsByStatus (EntryStatus.PENDING,
                                             EntryStatus.APPROVED, 
                                             EntryStatus.LOCALSUBMIT, filter_name=filter_name)
        return self.paginateResults (requests, page=page, entriesPerPage = self.cfg.entriesPerPage) 
    
    @templateXML
    @clear_cache (WorkflowEntry, Location)
    def getOperatorRequests (self, page=0, filter_name="", **args):
        token = SecurityToken ()
        securityApi = self.context.SecurityDBApi ()
        regions = securityApi.groupsForRole (token.dn, "Production Operator")
        requests = []
        for region in regions.values ():
            locationSelect = Location.select (Location.q.user_group == region["id"])
            if not locationSelect.count ():
                self.context.Logger ().message ("User %s is member of invalid location id %s" % (token.dn, region["id"]))
                continue
            
            location = locationSelect[0]
            self.context.Logger ().message (location.name)
            for request in location.entries:
                if request.status in (EntryStatus.APPROVED, EntryStatus.DONE, EntryStatus.RUNNING):
                    if filter_name in "" or filter_name in request.primary_dataset:
                        requests.append (request)
        
        for request in Location.get (1).entries:
            if request.status in (EntryStatus.APPROVED, EntryStatus.DONE, EntryStatus.RUNNING):
                if filter_name in ""  or filter_name in request.primary_dataset:
                    requests.append (request)
        
        return self.paginateResults (requests, page=page, entriesPerPage = self.cfg.entriesPerPage)
        
    def __createRequestSpec (self, entry, approvalId=-1):
        reqReleaseName = entry.project.name
        securityApi = self.context.SecurityDBApi ()
        try:
            user = securityApi.getUsernameFromID (entry.workflow.owner_id)
        except:
            user = "Unknown"
        prodcommonPlugin = self.prodCommonPlugins[entry.prodcommon_version]
        workflowspecs = [prodcommonPlugin.createRequestSpec (entry.id, cmsswVersion=entry.project.name)]
        return {"name": entry.name,
                "user": user,
                "id": entry.id,
                "approval_id": approvalId,
                "workflow_id": entry.workflowID,
                "events": entry.requested_events,
                "workflowspecs": workflowspecs}
        

    # TODO: add an option to limit the amount of data transferred. Right now capped to 10.
    # TODO: merge the Workflow creation stuff in one single function to be used by both ProdRequest and ProdRequestPA.
    @templateXML
    @clear_cache (WorkflowEntry, Payload, ApprovedRequest)
    @require_args ("lastSeen", onFail=list)
    def getRequestSpecsSlice (self, lastSeen, consumer="PRODMANAGER", **args):
        debug = args.has_key ("debug") or False
        consumerId = getattr (ConsumerName, consumer)
        approvedRequests = ApprovedRequest.select (AND (ApprovedRequest.q.id > lastSeen,
                                                        ApprovedRequest.q.consumer == consumerId))
        requests = [ (self.__createRequestSpec (approvedRequest.entry, approvedRequest.id), 
                      approvedRequest.entry)
                     for approvedRequest in approvedRequests[0:10] ]
        return {'requests': requests}
    
    @templateXML
    @clear_cache (WorkflowEntry, Payload, ApprovedRequest)
    @require_args ("request_id", onFail=dict)
    def getRequestSpec (self, request_id, **args):
        return self.__createRequestSpec (WorkflowEntry.get (request_id))
        
    @xmlBooleanResponse
    @require_args ("editedWorkflow", "name", onFail=AlwaysFalse)
    @is_authenticated (onFail=AlwaysFalse)
    def renameWorkflow (self, editedWorkflow, name, **args):
        self.context.Logger ().message ("****** Changing name of the workflow %s to: %s" % (Workflow.get (int (editedWorkflow)) , name))
        Workflow.get (int (editedWorkflow)).name = name
        return True

    @templateXML
    def getPendingRequests (self, **args):
        pendingRequests = self.getRequestsByStatus (EntryStatus.PENDING)
        page=int (args["page"])
        initialRequest = self.cfg.entriesPerPage * page
        finalRequest = self.cfg.entriesPerPage * (page + 1)
        if len (pendingRequests) < finalRequest:
            finalRequest = len (pendingRequests)
        return {"initialRequest": initialRequest,
                "finalRequest": finalRequest,
                "totalRequests": len(pendingRequests) ,
                "requests": pendingRequests[initialRequest:finalRequest]}

    @exposeSerialized (serializer=PythonListSerializer ('entries', 'entry'))
    @clear_cache (WorkflowEntry, Payload)
    @require_args ("workflow", onFail=list)
    def getWorkflowBrowserEntries (self, workflow, **args):
        returnEntries = []
        for entry in Workflow.get (workflow).entries:
            returnEntries.append ({"name": entry.payload.name, 
                                   "id": entry.id, 
                                   "primary_dataset": entry.primary_dataset,
                                   "output_dataset": entry.output_dataset,
                                   "pileup_dataset": entry.pileup_dataset,
                                   "selection_efficiency": entry.selection_efficiency,
                                   "cross_section": entry.cross_section,
                                   "events": entry.requested_events,
                                   "physics_group": IdToPhysicsGroup(entry.physics_group),
                                   "cmssw_version": entry.project.name})
        return returnEntries

    # TODO: Do a better query for the payload informations.
    @exposeSerialized (serializer=PythonListSerializer ("entries", "entry"))
    @clear_cache (WorkflowEntry, Payload)
    @ref_workflow_by_id
    @require_args ("workflow", onFail=list)
    def getWorkflowEntries (self, workflow, **args):
        entries = []
        for entry in workflow.entries:
            entries.append ({"name": entry.payload.name, "id": entry.payload.id}) 
        return entries
    
    @exposeSerialized (serializer=SQLObjectSerializer ("workflows", "workflow", exclusionList=["owner_id", "creation_time", "status"]))
    @clear_cache (Workflow)
    def getAvailableWorkflows (self, **args):
        token = SecurityToken ()
        securityApi = self.context.SecurityDBApi ()
        if not len (securityApi.groupsForRole (token.dn, "Production Manager")):
            workflowConditions = OR (Workflow.q.owner_id == 0, Workflow.q.owner_id == SecurityToken().userId)
            workflowSelect = Workflow.select (AND (workflowConditions, Workflow.q.name != "Dummy name"))
        else:
            workflowSelect = Workflow.select (Workflow.q.name != "Dummy name")
        return workflowSelect
        
    @templateXML
    @ref_payload_by_id ('payload')
    @clear_cache (Payload)
    def getPayload (self, payload=None, **args):
        if args.has_key ("collapsed"):
            parameterSet = ""
        elif args.has_key ("short"):
            #FIXME: shorten the parameter set.
            lines = payload.parameterset.split ("\n")[0:4];
            lines.append ("...")
            parameterSet = "<br/>".join (lines)
        else:
            parameterSet = "<br/>".join (payload.parameterset.split ("\n"))
        return {'id': payload.id,
                'name': payload.name,
                'parameterSet': "<![CDATA[" + parameterSet + "]]>"}
                
    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    @ref_workflow_entry_by_id
    @check_ownership (FetchFromArgs ("entry"), ModifiableRequest ())  
    @require_args ("entry", "dataset_name")
    @catch_exception (SQLObjectNotFound)
    def setWorkflowEntryDataset (self, datasetType, entry=None, dataset_name=None, **args):
        if datasetType not in ["primary", "pileup", "input"]:
            return False
        setattr (entry, datasetType + "_dataset", dataset_name)
        entry.spec_info = None
        return True

    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    @ref_workflow_entry_by_id
    @require_args ("entry", "key", "value", "value_type", onFail=AlwaysFalse)
    @check_ownership (FetchFromArgs ("entry"), ModifiableRequest ())
    @catch_exception (SQLObjectNotFound)
    def setWorkflowEntryProperty(self, key, value, entry=None, value_type="str", **args):
        if entry.status in [EntryStatus.INJECTED, EntryStatus.PENDING]:
            value = globals()["__builtins__"][value_type] (value)
            setattr (entry, key, value)
            entry.spec_info = None
            return True
        return False

    @xmlBooleanResponse
    @is_authenticated (onFail=bool)
    @ref_workflow_entry_by_id
    @ref_project_by_id
    @ref_project_by_name
    @require_args ('entry', 'project', onFail=bool)
    @check_ownership (FetchFromArgs ("entry"), ModifiableRequest ())
    @catch_exception (SQLObjectNotFound)
    def setWorkflowEntryCMSSWVersion (self, entry=None, project=None, **args):
        if entry.status in [EntryStatus.INJECTED, EntryStatus.PENDING]:
            entry.project = project
            print defaultProdCommon (project.name)
            entry.prodcommon_version = defaultProdCommon (project.name)
            entry.spec_info = None
            return True
        return False
    
    
    # TODO: this will kill only the entry but not the associated payload. Some reference counting is needed for payloads.
    # Or (better) temporary workflows should be done in a separate table and then committed to the big one when saved or submitted.
    @xmlBooleanResponse
    @is_authenticated (onFail=AlwaysFalse)
    @ref_workflow_entry_by_id
    @require_args ("entry")
    @check_ownership (FetchFromArgs ("entry"), ModifiableRequest ())  
    @catch_exception (SQLObjectNotFound)
    def destroyWorkflowEntry (self, entry=None, **args):
        entry.destroySelf ()
        return True
    
    @templatepage
    def requestStatusPage (self, **args):
        requestsArgs = selectFromDict (args, "request_id")
        return {"request_id": requestsArgs["request_id"]}
    
    def getStatusConditions (self):
        conditions = {}
        conditions["pending"] = [WorkflowEntry.q.status == EntryStatus.PENDING]
        conditions["approved"] = [WorkflowEntry.q.status == EntryStatus.APPROVED]
        conditions['running'] = [WorkflowEntry.q.status == EntryStatus.RUNNING]
        conditions['done'] = [WorkflowEntry.q.status == EntryStatus.DONE]
        return conditions
    
    def getUserProductionStatus (self):
        productionStatus = {}
        for key, condition in self.getStatusConditions ().items ():
            condition.append (Workflow.q.id == WorkflowEntry.q.workflowID)
            condition.append (Workflow.q.owner_id == SecurityToken ().userId)    
            productionStatus[key] = WorkflowEntry.select (AND (*condition)).count ()
        return productionStatus
                
    # TODO: good enough for now but needs to implement the "completed" bit.
    @templatepage
    @clear_cache (WorkflowEntry)
    def getHome (self, **args):
        userStatus = self.getUserProductionStatus ()
        assignedRequests = 0
        waitingApproval = 0
        token = SecurityToken ()
        securityApi = self.context.SecurityDBApi ()
        regions = securityApi.groupsForRole (token.dn, "Production Operator")
        if len (regions):
            self.context.Logger ().message ("Groups: %s" % regions)
            regionsCondition = []
            for region in regions.values ():
                locations = Location.select (Location.q.user_group == region["id"])
                if locations.count ():
                    entries = locations[0].entries
                    for entry in entries:
                        if entry.status == EntryStatus.APPROVED or entry.status == EntryStatus.APPROVED_PRODMANAGER:
                            assignedRequests += 1
                    self.context.Logger ().message ("Group: %s" % region)
        else:
            self.context.Logger ().message ("User is not a Production Operator.")
        isProdManager = self.context.SecurityDBApi ().hasGroupResponsibility (SecurityToken ().dn, "Production Manager", "ProdRequest")
        isAdmin = self.context.SecurityDBApi ().hasGroupResponsibility (SecurityToken ().dn, "Admin", "ProdRequest")
        
        if isAdmin or isProdManager:
            waitingApproval = globalStatus["pending"]
        
        return {'globalStatus': self.monitoring.getGlobalProductionStatus (),
                'userStatus': userStatus,
                'waitingApproval': waitingApproval,
                'assignedRequests': 0}
    
    @templatepage
    def gotoLogin (self, **args):
        return {'requesterPage': args["requesterPage"]}
    
    @templatepage
    def redirect (self, *args, **kw):
        return {"redirect": '../Studio/login'}
    
    @expose
    @require_args ("dbsUrl", "dbsInstance")
    def changeConfiguration (self, dbsUrl, dbsInstance, *args, **kw):
        try:
            self.context.ProdRequestDBSCore ().initDBSCgiApi (dbsUrl, dbsInstance)
            pass
        except KeyError:
            print "ProdRequestDBS not found"
        return self.templatePage ("redirect", {'redirect': "../ProdRequest/getAdministrationPage"})

    @xmlBooleanResponse
    @ref_workflow_entry_by_id
    @declare_arg_converter ('value', '_text', lambda value_text : getattr (EntryStatus, value_text))
    @declare_arg_converter ('value', '', lambda value : int (value))
    @require_args ("entry", "value", onFail=AlwaysFalse)
    @check_ownership (FetchFromArgs ("entry"), OperatorsRequest ())
    def changeEntryStatus (self, entry=None, value=None, **args):
        if entry.status == EntryStatus.APPROVED_PRODMANAGER:
            return False
        entry.status = value
        if value == EntryStatus.APPROVED_PRODMANAGER:
            ApprovedRequest (entry=entry, consumer=ConsumerName.PRODMANAGER)
        return True
    
    @expose
    def gotoMonitoringPage (self, **args):
        raise cherrypy.HTTPRedirect ("../ProdRequest/getMonitoringPage")

    @templatepage
    def getAllRequestsPage (self, **args):
        return {}
        
    @xmlBooleanResponse
    @ref_workflow_entry_by_id
    @ref_location_by_id
    @require_args ("entry", "location")
    @check_ownership (FetchFromArgs ("entry"), ModifiableRequest ())
    @catch_exception (SQLObjectNotFound)
    def setWorkflowEntryLocation (self, entry=None, location=None, **args):
        for oldLocation in entry.location_hints:
            entry.removeLocation (oldLocation)
        entry.addLocation (location)
        entry.spec_info = None
        return True

    def __queryRequests (self, *args, **kwds):
        queryStatements = []
        for x in args:
            key, value = x.split ("=")
            try:
                value, operator = value.split ("@")
            except:
                operator = "is"
            if key == "cmssw_version":
                projects = Project.select (Project.q.name == value)
                if not projects.count ():
                    return []
                key="projectID"
                value=projects[0].id
            
            if key == "status_text":
                key = "status"
                value = getattr (EntryStatus, value)
            
            if operator == "is":
                queryStatements.append (getattr (WorkflowEntry.q, key) == value)
            elif operator == "contains":
                queryStatements.append (LIKE(getattr (WorkflowEntry.q, key), "%%%s%%" % value))
            elif operator == "startswith":
                queryStatements.append (LIKE(getattr (WorkflowEntry.q, key), "%s%%" % value))
            elif operator == "endswith":
                queryStatements.append (LIKE(getattr (WorkflowEntry.q, key), "%%%s" % value))
        if kwds["queryType"] == "or":
            entries = WorkflowEntry.select (OR (*queryStatements))
        else:
            entries = WorkflowEntry.select (AND (*queryStatements))
        return entries

    @exposeSerialized (serializer=PythonListSerializer ("request", "request"))
    def queryRequests (self, *args, **kwds):
        entries = self.__queryRequests (*args, **kwds)
        results = [ {"request_id": entry.id}
                    for entry in entries]
        return results
    
    @exposeSerialized (serializer=PythonListSerializer ("groups", "group"))
    def getPhysicsGroups (self, **args):
        result = [ {"id": value, "label": key.replace("___", "/").replace ("__", "-")}
                   for key,value in PhysicsGroup._enum__dict.items () ]
        def sortById (x, y):
            return x["id"] - y["id"]
        result.sort (sortById)
        return result
        
    @xmlBooleanResponse
    @ref_workflow_entry_by_id
    @require_args ("entry")
    @check_ownership (FetchFromArgs ("entry"), ModifiableRequest ())
    @catch_exception (SQLObjectNotFound)
    def deleteRequest (self, entry=None, **kwds):
        if entry.status not in [EntryStatus.PENDING, EntryStatus.INJECTED]:
            return False
        token = SecurityToken ()
        if token.userId == entry.workflow.owner_id:
            entry.status = EntryStatus.RETIRED
            return True
        return False
    
    @xmlBooleanResponse
    @is_authorized (Role ("Production Manager"), Group ("ProdRequest"), onFail=bool)
    def emptyTrash (self):
        entries = WorkflowEntry.select (WorkflowEntry.q.status == EntryStatus.RETIRED)
        for entry in entries:
            entry.destroySelf ()
        return True
    
    @templateXML
    @require_args ("primary_dataset", "id", onFail=dict)
    def getRequestsByDataset (self, id, primary_dataset, **args):
        requests = WorkflowEntry.select (WorkflowEntry.q.primary_dataset == primary_dataset)
        return {"requests": requests, "call_id": id}
        
DeclarePlugin ("/Controllers/ProdRequest/Core", ProdRequest, {"baseUrl": "/ProdRequest"})

FrameworkApplication (label="ProdRequest",
                      name="ProdRequest",
                      url="../ProdRequest/initSession",
                      image="../Studio/images/mini.gif")

