from sqlobject import *
from Models import prodRequestDb
from Modules.Utilities.Language import enum

EntryPriorityHint = enum ("LOW", "NORMAL", "HIGH")
EntryStatus = enum ("INJECTED", "PENDING", "APPROVED", "FETCHED", "LOCALSUBMIT", "REJECTED", "RUNNING", "DONE", "RETIRED", "APPROVED_PRODMANAGER")
PhysicsGroup = enum ("Individual", "Higgs", "SUSYBSM", "EWK", "Top", "QCD", 
    "Diffraction", "OnSel", "Bphys", "Muons", "Egamma", 
    "JetMet", "PFlowTau", "Btag")
ConsumerName = enum ("OPERATOR", "PRODMANAGER")

g_IdToPhysicsGroup = {}
g_IdToEntryStatus = {}

for key, value in PhysicsGroup._enum__dict.items ():
    g_IdToPhysicsGroup[value] = key.replace ("___", "/").replace ("__", "-")

for key, value in EntryStatus._enum__dict.items ():
    g_IdToEntryStatus[value] = key

def IdToPhysicsGroup (id):
    return g_IdToPhysicsGroup[id]

def IdToEntryStatus (id):
    return g_IdToEntryStatus[id]


class WorkflowEntry (SQLObject):
    _connection = prodRequestDb.conn
    name = StringCol()
    workflow = ForeignKey ('Workflow')
    payload = ForeignKey ('Payload')
    payloadprev = ForeignKey ('Payload', default=None)
    payloadparent = ForeignKey ('Payload', default=None)
    primary_dataset = StringCol ()
    input_dataset = StringCol (default=None)
    output_dataset = StringCol (default=None)
    pileup_dataset = StringCol (default=None)
    status = IntCol ()
    priority_hint=IntCol (default=50)
    location_hints=RelatedJoin ('Location')
    requested_events=IntCol (default=1000)
    project=ForeignKey ('Project', default=None)
    selection_efficiency=FloatCol (default=1.0)
    cross_section=FloatCol (default=0.0)
    architecture=ForeignKey ('Architecture', default=None)
    physics_group=IntCol (default=PhysicsGroup.Individual)
    owner_id=IntCol (default=0)
    prodcommon_version=StringCol (default="PRODCOMMON_0_3_X")
    request_spec=StringCol (default="")
    generator_info=ForeignKey ('GeneratorInfo', default=None)
    preferred_consumer=IntCol (default=ConsumerName.OPERATOR)
    spec_info=ForeignKey ('SpecInfo', default=None)